@@ -19,12 +19,9 @@
 #include <inttypes.h>
 
 #include <functional>
-#include <initializer_list>
 #include <memory>
-#include <string>
 #include <utility>
 
-#include "absl/meta/type_traits.h"
 #include "absl/status/status.h"
 #include "absl/strings/str_cat.h"
 #include "absl/strings/str_format.h"
@@ -32,10 +29,12 @@
 
 #include <grpc/compression.h>
 #include <grpc/grpc.h>
+#include <grpc/impl/channel_arg_names.h>
 #include <grpc/impl/compression_types.h>
 #include <grpc/support/log.h>
 
 #include "src/core/ext/filters/message_size/message_size_filter.h"
+#include "src/core/lib/channel/call_tracer.h"
 #include "src/core/lib/channel/channel_args.h"
 #include "src/core/lib/channel/channel_stack.h"
 #include "src/core/lib/channel/context.h"
@@ -47,8 +46,7 @@
 #include "src/core/lib/promise/context.h"
 #include "src/core/lib/promise/latch.h"
 #include "src/core/lib/promise/pipe.h"
-#include "src/core/lib/promise/poll.h"
-#include "src/core/lib/promise/race.h"
+#include "src/core/lib/promise/prioritized_race.h"
 #include "src/core/lib/resource_quota/arena.h"
 #include "src/core/lib/slice/slice_buffer.h"
 #include "src/core/lib/surface/call.h"
@@ -58,6 +56,11 @@
 
 namespace grpc_core {
 
+const NoInterceptor ServerCompressionFilter::Call::OnServerTrailingMetadata;
+const NoInterceptor ServerCompressionFilter::Call::OnFinalize;
+const NoInterceptor ClientCompressionFilter::Call::OnServerTrailingMetadata;
+const NoInterceptor ClientCompressionFilter::Call::OnFinalize;
+
 const grpc_channel_filter ClientCompressionFilter::kFilter =
     MakePromiseBasedFilter<ClientCompressionFilter, FilterEndpoint::kClient,
                            kFilterExaminesServerInitialMetadata |
@@ -79,7 +82,7 @@ absl::StatusOr<ServerCompressionFilter> ServerCompressionFilter::Create(
   return ServerCompressionFilter(args);
 }
 
-CompressionFilter::CompressionFilter(const ChannelArgs& args)
+ChannelCompression::ChannelCompression(const ChannelArgs& args)
     : max_recv_size_(GetMaxRecvSizeFromChannelArgs(args)),
       message_size_service_config_parser_index_(
           MessageSizeParser::ParserIndex()),
@@ -107,12 +110,18 @@ CompressionFilter::CompressionFilter(const ChannelArgs& args)
   }
 }
 
-MessageHandle CompressionFilter::CompressMessage(
+MessageHandle ChannelCompression::CompressMessage(
     MessageHandle message, grpc_compression_algorithm algorithm) const {
   if (GRPC_TRACE_FLAG_ENABLED(grpc_compression_trace)) {
-    gpr_log(GPR_ERROR, "CompressMessage: len=%" PRIdPTR " alg=%d flags=%d",
+    gpr_log(GPR_INFO, "CompressMessage: len=%" PRIdPTR " alg=%d flags=%d",
             message->payload()->Length(), algorithm, message->flags());
   }
+  auto* call_context = GetContext<grpc_call_context_element>();
+  auto* call_tracer = static_cast<CallTracerInterface*>(
+      call_context[GRPC_CONTEXT_CALL_TRACER].value);
+  if (call_tracer != nullptr) {
+    call_tracer->RecordSendMessage(*message->payload());
+  }
   // Check if we're allowed to compress this message
   // (apps might want to disable compression for certain messages to avoid
   // crime/beast like vulns).
@@ -143,6 +152,9 @@ MessageHandle CompressionFilter::CompressMessage(
     }
     tmp.Swap(payload);
     flags |= GRPC_WRITE_INTERNAL_COMPRESS;
+    if (call_tracer != nullptr) {
+      call_tracer->RecordSendCompressedMessage(*message->payload());
+    }
   } else {
     if (GRPC_TRACE_FLAG_ENABLED(grpc_compression_trace)) {
       const char* algo_name;
@@ -156,20 +168,27 @@ MessageHandle CompressionFilter::CompressMessage(
   return message;
 }
 
-absl::StatusOr<MessageHandle> CompressionFilter::DecompressMessage(
-    MessageHandle message, DecompressArgs args) const {
+absl::StatusOr<MessageHandle> ChannelCompression::DecompressMessage(
+    bool is_client, MessageHandle message, DecompressArgs args) const {
   if (GRPC_TRACE_FLAG_ENABLED(grpc_compression_trace)) {
-    gpr_log(GPR_ERROR, "DecompressMessage: len=%" PRIdPTR " max=%d alg=%d",
+    gpr_log(GPR_INFO, "DecompressMessage: len=%" PRIdPTR " max=%d alg=%d",
             message->payload()->Length(),
             args.max_recv_message_length.value_or(-1), args.algorithm);
   }
+  auto* call_context = GetContext<grpc_call_context_element>();
+  auto* call_tracer = static_cast<CallTracerInterface*>(
+      call_context[GRPC_CONTEXT_CALL_TRACER].value);
+  if (call_tracer != nullptr) {
+    call_tracer->RecordReceivedMessage(*message->payload());
+  }
   // Check max message length.
   if (args.max_recv_message_length.has_value() &&
       message->payload()->Length() >
           static_cast<size_t>(*args.max_recv_message_length)) {
     return absl::ResourceExhaustedError(absl::StrFormat(
-        "Received message larger than max (%u vs. %d)",
-        message->payload()->Length(), *args.max_recv_message_length));
+        "%s: Received message larger than max (%u vs. %d)",
+        is_client ? "CLIENT" : "SERVER", message->payload()->Length(),
+        *args.max_recv_message_length));
   }
   // Check if decompression is enabled (if not, we can just pass the message
   // up).
@@ -189,10 +208,13 @@ absl::StatusOr<MessageHandle> CompressionFilter::DecompressMessage(
   message->payload()->Swap(&decompressed_slices);
   message->mutable_flags() &= ~GRPC_WRITE_INTERNAL_COMPRESS;
   message->mutable_flags() |= GRPC_WRITE_INTERNAL_TEST_ONLY_WAS_COMPRESSED;
+  if (call_tracer != nullptr) {
+    call_tracer->RecordReceivedDecompressedMessage(*message->payload());
+  }
   return std::move(message);
 }
 
-grpc_compression_algorithm CompressionFilter::HandleOutgoingMetadata(
+grpc_compression_algorithm ChannelCompression::HandleOutgoingMetadata(
     grpc_metadata_batch& outgoing_metadata) {
   const auto algorithm = outgoing_metadata.Take(GrpcInternalEncodingRequest())
                              .value_or(default_compression_algorithm());
@@ -205,7 +227,7 @@ grpc_compression_algorithm CompressionFilter::HandleOutgoingMetadata(
   return algorithm;
 }
 
-CompressionFilter::DecompressArgs CompressionFilter::HandleIncomingMetadata(
+ChannelCompression::DecompressArgs ChannelCompression::HandleIncomingMetadata(
     const grpc_metadata_batch& incoming_metadata) {
   // Configure max receive size.
   auto max_recv_message_length = max_recv_size_;
@@ -216,92 +238,59 @@ CompressionFilter::DecompressArgs CompressionFilter::HandleIncomingMetadata(
   if (limits != nullptr && limits->max_recv_size().has_value() &&
       (!max_recv_message_length.has_value() ||
        *limits->max_recv_size() < *max_recv_message_length)) {
-    max_recv_message_length = *limits->max_recv_size();
+    max_recv_message_length = limits->max_recv_size();
   }
   return DecompressArgs{incoming_metadata.get(GrpcEncodingMetadata())
                             .value_or(GRPC_COMPRESS_NONE),
                         max_recv_message_length};
 }
 
-ArenaPromise<ServerMetadataHandle> ClientCompressionFilter::MakeCallPromise(
-    CallArgs call_args, NextPromiseFactory next_promise_factory) {
-  auto compression_algorithm =
-      HandleOutgoingMetadata(*call_args.client_initial_metadata);
-  call_args.client_to_server_messages->InterceptAndMap(
-      [compression_algorithm,
-       this](MessageHandle message) -> absl::optional<MessageHandle> {
-        return CompressMessage(std::move(message), compression_algorithm);
-      });
-  auto* decompress_args = GetContext<Arena>()->New<DecompressArgs>(
-      DecompressArgs{GRPC_COMPRESS_NONE, absl::nullopt});
-  auto* decompress_err =
-      GetContext<Arena>()->New<Latch<ServerMetadataHandle>>();
-  call_args.server_initial_metadata->InterceptAndMap(
-      [decompress_args, this](ServerMetadataHandle server_initial_metadata)
-          -> absl::optional<ServerMetadataHandle> {
-        if (server_initial_metadata == nullptr) return absl::nullopt;
-        *decompress_args = HandleIncomingMetadata(*server_initial_metadata);
-        return std::move(server_initial_metadata);
-      });
-  call_args.server_to_client_messages->InterceptAndMap(
-      [decompress_err, decompress_args,
-       this](MessageHandle message) -> absl::optional<MessageHandle> {
-        auto r = DecompressMessage(std::move(message), *decompress_args);
-        if (!r.ok()) {
-          decompress_err->Set(ServerMetadataFromStatus(r.status()));
-          return absl::nullopt;
-        }
-        return std::move(*r);
-      });
-  // Run the next filter, and race it with getting an error from decompression.
-  return Race(next_promise_factory(std::move(call_args)),
-              decompress_err->Wait());
+void ClientCompressionFilter::Call::OnClientInitialMetadata(
+    ClientMetadata& md, ClientCompressionFilter* filter) {
+  compression_algorithm_ =
+      filter->compression_engine_.HandleOutgoingMetadata(md);
 }
 
-ArenaPromise<ServerMetadataHandle> ServerCompressionFilter::MakeCallPromise(
-    CallArgs call_args, NextPromiseFactory next_promise_factory) {
-  auto decompress_args =
-      HandleIncomingMetadata(*call_args.client_initial_metadata);
-  auto* decompress_err =
-      GetContext<Arena>()->New<Latch<ServerMetadataHandle>>();
-  call_args.client_to_server_messages->InterceptAndMap(
-      [decompress_err, decompress_args,
-       this](MessageHandle message) -> absl::optional<MessageHandle> {
-        auto r = DecompressMessage(std::move(message), decompress_args);
-        if (grpc_call_trace.enabled()) {
-          gpr_log(GPR_DEBUG, "DecompressMessage returned %s",
-                  r.status().ToString().c_str());
-        }
-        if (!r.ok()) {
-          decompress_err->Set(ServerMetadataFromStatus(r.status()));
-          return absl::nullopt;
-        }
-        return std::move(*r);
-      });
-  auto* compression_algorithm =
-      GetContext<Arena>()->New<grpc_compression_algorithm>();
-  call_args.server_initial_metadata->InterceptAndMap(
-      [this, compression_algorithm](ServerMetadataHandle md) {
-        if (grpc_call_trace.enabled()) {
-          gpr_log(GPR_INFO, "%s[compression] Write metadata",
-                  Activity::current()->DebugTag().c_str());
-        }
-        // Find the compression algorithm.
-        *compression_algorithm = HandleOutgoingMetadata(*md);
-        return md;
-      });
-  call_args.server_to_client_messages->InterceptAndMap(
-      [compression_algorithm,
-       this](MessageHandle message) -> absl::optional<MessageHandle> {
-        return CompressMessage(std::move(message), *compression_algorithm);
-      });
-  // Concurrently:
-  // - call the next filter
-  // - decompress incoming messages
-  // - wait for initial metadata to be sent, and then commence compression of
-  //   outgoing messages
-  return Race(next_promise_factory(std::move(call_args)),
-              decompress_err->Wait());
+MessageHandle ClientCompressionFilter::Call::OnClientToServerMessage(
+    MessageHandle message, ClientCompressionFilter* filter) {
+  return filter->compression_engine_.CompressMessage(std::move(message),
+                                                     compression_algorithm_);
+}
+
+void ClientCompressionFilter::Call::OnServerInitialMetadata(
+    ServerMetadata& md, ClientCompressionFilter* filter) {
+  decompress_args_ = filter->compression_engine_.HandleIncomingMetadata(md);
+}
+
+absl::StatusOr<MessageHandle>
+ClientCompressionFilter::Call::OnServerToClientMessage(
+    MessageHandle message, ClientCompressionFilter* filter) {
+  return filter->compression_engine_.DecompressMessage(
+      /*is_client=*/true, std::move(message), decompress_args_);
+}
+
+void ServerCompressionFilter::Call::OnClientInitialMetadata(
+    ClientMetadata& md, ServerCompressionFilter* filter) {
+  decompress_args_ = filter->compression_engine_.HandleIncomingMetadata(md);
+}
+
+absl::StatusOr<MessageHandle>
+ServerCompressionFilter::Call::OnClientToServerMessage(
+    MessageHandle message, ServerCompressionFilter* filter) {
+  return filter->compression_engine_.DecompressMessage(
+      /*is_client=*/false, std::move(message), decompress_args_);
+}
+
+void ServerCompressionFilter::Call::OnServerInitialMetadata(
+    ServerMetadata& md, ServerCompressionFilter* filter) {
+  compression_algorithm_ =
+      filter->compression_engine_.HandleOutgoingMetadata(md);
+}
+
+MessageHandle ServerCompressionFilter::Call::OnServerToClientMessage(
+    MessageHandle message, ServerCompressionFilter* filter) {
+  return filter->compression_engine_.CompressMessage(std::move(message),
+                                                     compression_algorithm_);
 }
 
 }  // namespace grpc_core
