@@ -37,6 +37,7 @@
 #include "src/core/lib/channel/channel_stack_builder.h"
 #include "src/core/lib/gprpp/debug_location.h"
 #include "src/core/lib/surface/channel_stack_type.h"
+#include "src/core/lib/transport/call_filters.h"
 
 /// This module provides a way for plugins (and the grpc core library itself)
 /// to register mutators for channel stacks.
@@ -76,11 +77,26 @@ class ChannelInit {
     kCount
   };
 
+  // Vtable-like data structure for channel data initialization
+  struct ChannelFilterVtable {
+    size_t size;
+    size_t alignment;
+    absl::Status (*init)(void* data, const ChannelArgs& args);
+    void (*destroy)(void* data);
+    void (*add_to_stack_builder)(void* data,
+                                 CallFilters::StackBuilder& builder);
+  };
+
   class FilterRegistration {
    public:
+    // TODO(ctiller): Remove grpc_channel_filter* arg when that can be
+    // deprecated (once filter stack is removed).
     explicit FilterRegistration(const grpc_channel_filter* filter,
+                                const ChannelFilterVtable* vtable,
                                 SourceLocation registration_source)
-        : filter_(filter), registration_source_(registration_source) {}
+        : filter_(filter),
+          vtable_(vtable),
+          registration_source_(registration_source) {}
     FilterRegistration(const FilterRegistration&) = delete;
     FilterRegistration& operator=(const FilterRegistration&) = delete;
 
@@ -146,15 +162,21 @@ class ChannelInit {
     // Add a predicate that ensures this filter does not appear in the minimal
     // stack.
     FilterRegistration& ExcludeFromMinimalStack();
+    FilterRegistration& SkipV3() {
+      skip_v3_ = true;
+      return *this;
+    }
 
    private:
     friend class ChannelInit;
     const grpc_channel_filter* const filter_;
+    const ChannelFilterVtable* const vtable_;
     std::vector<const grpc_channel_filter*> after_;
     std::vector<const grpc_channel_filter*> before_;
     std::vector<InclusionPredicate> predicates_;
     bool terminal_ = false;
     bool before_all_ = false;
+    bool skip_v3_ = false;
     SourceLocation registration_source_;
   };
 
@@ -166,13 +188,25 @@ class ChannelInit {
     // properties of the filter being registered.
     // TODO(ctiller): remove in favor of the version that does not mention
     // grpc_channel_filter
-    FilterRegistration& RegisterFilter(grpc_channel_stack_type type,
-                                       const grpc_channel_filter* filter,
-                                       SourceLocation registration_source = {});
+    FilterRegistration& RegisterFilter(
+        grpc_channel_stack_type type, const grpc_channel_filter* filter,
+        const ChannelFilterVtable* vtable = nullptr,
+        SourceLocation registration_source = {});
     template <typename Filter>
     FilterRegistration& RegisterFilter(
         grpc_channel_stack_type type, SourceLocation registration_source = {}) {
-      return RegisterFilter(type, &Filter::kFilter, registration_source);
+      return RegisterFilter(type, &Filter::kFilter,
+                            VtableForType<Filter>::vtable(),
+                            registration_source);
+    }
+
+    // Filter does not participate in v3
+    template <typename Filter>
+    FilterRegistration& RegisterV2Filter(
+        grpc_channel_stack_type type, SourceLocation registration_source = {}) {
+      return RegisterFilter(type, &Filter::kFilter, nullptr,
+                            registration_source)
+          .SkipV3();
     }
 
     // Register a post processor for the builder.
@@ -198,22 +232,74 @@ class ChannelInit {
                                   [static_cast<int>(PostProcessorSlot::kCount)];
   };
 
+  // A set of channel filters that can be added to a call stack.
+  // TODO(ctiller): move this out so it can be used independently of
+  // the global registration mechanisms.
+  class StackSegment final {
+   public:
+    // Registration of one channel filter in the stack.
+    struct ChannelFilter {
+      size_t offset;
+      const ChannelFilterVtable* vtable;
+    };
+
+    StackSegment() = default;
+    explicit StackSegment(std::vector<ChannelFilter> filters,
+                          uint8_t* channel_data);
+    StackSegment(const StackSegment& other) = delete;
+    StackSegment& operator=(const StackSegment& other) = delete;
+    StackSegment(StackSegment&& other) noexcept = default;
+    StackSegment& operator=(StackSegment&& other) = default;
+
+    // Add this segment to a call filter stack builder
+    void AddToCallFilterStack(CallFilters::StackBuilder& builder);
+
+   private:
+    // Combined channel data for the stack
+    class ChannelData : public RefCounted<ChannelData> {
+     public:
+      explicit ChannelData(std::vector<ChannelFilter> filters,
+                           uint8_t* channel_data);
+      ~ChannelData() override;
+
+      void AddToCallFilterStack(CallFilters::StackBuilder& builder);
+
+     private:
+      std::vector<ChannelFilter> filters_;
+      uint8_t* channel_data_;
+    };
+
+    RefCountedPtr<ChannelData> data_;
+  };
+
   /// Construct a channel stack of some sort: see channel_stack.h for details
   /// \a builder is the channel stack builder to build into.
   GRPC_MUST_USE_RESULT
   bool CreateStack(ChannelStackBuilder* builder) const;
 
+  // Create a segment of a channel stack.
+  // Terminators and post processors are not included in this construction:
+  // terminators are a legacy filter-stack concept, and post processors
+  // need to migrate to other mechanisms.
+  // TODO(ctiller): figure out other mechanisms.
+  absl::StatusOr<StackSegment> CreateStackSegment(
+      grpc_channel_stack_type type, const ChannelArgs& args) const;
+
  private:
   struct Filter {
-    Filter(const grpc_channel_filter* filter,
-           std::vector<InclusionPredicate> predicates,
+    Filter(const grpc_channel_filter* filter, const ChannelFilterVtable* vtable,
+           std::vector<InclusionPredicate> predicates, bool skip_v3,
            SourceLocation registration_source)
         : filter(filter),
+          vtable(vtable),
           predicates(std::move(predicates)),
-          registration_source(registration_source) {}
+          registration_source(registration_source),
+          skip_v3(skip_v3) {}
     const grpc_channel_filter* filter;
+    const ChannelFilterVtable* vtable;
     std::vector<InclusionPredicate> predicates;
     SourceLocation registration_source;
+    bool skip_v3 = false;
     bool CheckPredicates(const ChannelArgs& args) const;
   };
   struct StackConfig {
@@ -221,6 +307,18 @@ class ChannelInit {
     std::vector<Filter> terminators;
     std::vector<PostProcessor> post_processors;
   };
+
+  template <typename T, typename = void>
+  struct VtableForType {
+    static const ChannelFilterVtable* vtable() { return nullptr; }
+  };
+
+  template <typename T>
+  struct VtableForType<T, absl::void_t<typename T::Call>> {
+    static const ChannelFilterVtable kVtable;
+    static const ChannelFilterVtable* vtable() { return &kVtable; }
+  };
+
   StackConfig stack_configs_[GRPC_NUM_CHANNEL_STACK_TYPES];
 
   static StackConfig BuildStackConfig(
@@ -228,6 +326,22 @@ class ChannelInit {
       PostProcessor* post_processors, grpc_channel_stack_type type);
 };
 
+template <typename T>
+const ChannelInit::ChannelFilterVtable
+    ChannelInit::VtableForType<T, absl::void_t<typename T::Call>>::kVtable = {
+        sizeof(T), alignof(T),
+        [](void* data, const ChannelArgs& args) -> absl::Status {
+          // TODO(ctiller): fill in ChannelFilter::Args (2nd arg)
+          absl::StatusOr<T> r = T::Create(args, {});
+          if (!r.ok()) return r.status();
+          new (data) T(std::move(*r));
+          return absl::OkStatus();
+        },
+        [](void* data) { static_cast<T*>(data)->~T(); },
+        [](void* data, CallFilters::StackBuilder& builder) {
+          builder.Add(static_cast<T*>(data));
+        }};
+
 }  // namespace grpc_core
 
 #endif  // GRPC_SRC_CORE_LIB_SURFACE_CHANNEL_INIT_H
