@@ -19,10 +19,14 @@
 #include <string.h>
 
 #include <algorithm>
+#include <string>
 
+#include "absl/base/no_destructor.h"
+#include "absl/container/flat_hash_set.h"
 #include "absl/strings/escaping.h"
 #include "absl/strings/match.h"
 #include "absl/strings/str_cat.h"
+#include "absl/strings/str_format.h"
 
 #include "src/core/lib/transport/timeout_encoding.h"
 
@@ -34,15 +38,73 @@ void DebugStringBuilder::Add(absl::string_view key, absl::string_view value) {
   absl::StrAppend(&out_, absl::CEscape(key), ": ", absl::CEscape(value));
 }
 
+void DebugStringBuilder::AddAfterRedaction(absl::string_view key,
+                                           absl::string_view value) {
+  if (IsAllowListed(key)) {
+    Add(key, value);
+  } else {
+    Add(key, absl::StrCat(value.size(), " bytes redacted by allow listing."));
+  }
+}
+
+bool DebugStringBuilder::IsAllowListed(const absl::string_view key) const {
+  static const absl::NoDestructor<absl::flat_hash_set<std::string>> allow_list(
+      [] {
+        absl::flat_hash_set<std::string> allow_list;
+        // go/keep-sorted start
+        allow_list.insert(std::string(ContentTypeMetadata::key()));
+        allow_list.insert(std::string(EndpointLoadMetricsBinMetadata::key()));
+        allow_list.insert(std::string(GrpcAcceptEncodingMetadata::key()));
+        allow_list.insert(std::string(GrpcEncodingMetadata::key()));
+        allow_list.insert(std::string(GrpcInternalEncodingRequest::key()));
+        allow_list.insert(std::string(GrpcLbClientStatsMetadata::key()));
+        allow_list.insert(std::string(GrpcMessageMetadata::key()));
+        allow_list.insert(std::string(GrpcPreviousRpcAttemptsMetadata::key()));
+        allow_list.insert(std::string(GrpcRetryPushbackMsMetadata::key()));
+        allow_list.insert(std::string(GrpcServerStatsBinMetadata::key()));
+        allow_list.insert(std::string(GrpcStatusMetadata::key()));
+        allow_list.insert(std::string(GrpcTagsBinMetadata::key()));
+        allow_list.insert(std::string(GrpcTimeoutMetadata::key()));
+        allow_list.insert(std::string(GrpcTraceBinMetadata::key()));
+        allow_list.insert(std::string(HostMetadata::key()));
+        allow_list.insert(std::string(HttpAuthorityMetadata::key()));
+        allow_list.insert(std::string(HttpMethodMetadata::key()));
+        allow_list.insert(std::string(HttpPathMetadata::key()));
+        allow_list.insert(std::string(HttpSchemeMetadata::key()));
+        allow_list.insert(std::string(HttpStatusMetadata::key()));
+        allow_list.insert(std::string(LbCostBinMetadata::key()));
+        allow_list.insert(std::string(LbTokenMetadata::key()));
+        allow_list.insert(std::string(TeMetadata::key()));
+        allow_list.insert(std::string(UserAgentMetadata::key()));
+        allow_list.insert(std::string(XEnvoyPeerMetadata::key()));
+
+        // go/keep-sorted end
+        // go/keep-sorted start
+        allow_list.insert(std::string(GrpcCallWasCancelled::DebugKey()));
+        allow_list.insert(std::string(GrpcRegisteredMethod::DebugKey()));
+        allow_list.insert(std::string(GrpcStatusContext::DebugKey()));
+        allow_list.insert(std::string(GrpcStatusFromWire::DebugKey()));
+        allow_list.insert(std::string(GrpcStreamNetworkState::DebugKey()));
+        allow_list.insert(std::string(GrpcTarPit::DebugKey()));
+        allow_list.insert(std::string(GrpcTrailersOnly::DebugKey()));
+        allow_list.insert(std::string(PeerString::DebugKey()));
+        allow_list.insert(std::string(WaitForReady::DebugKey()));
+        // go/keep-sorted end
+        return allow_list;
+      }());
+  return allow_list->contains(key);
+}
+
 void UnknownMap::Append(absl::string_view key, Slice value) {
-  unknown_.EmplaceBack(Slice::FromCopiedString(key), value.Ref());
+  unknown_.emplace_back(Slice::FromCopiedString(key), value.Ref());
 }
 
 void UnknownMap::Remove(absl::string_view key) {
-  unknown_.SetEnd(std::remove_if(unknown_.begin(), unknown_.end(),
-                                 [key](const std::pair<Slice, Slice>& p) {
-                                   return p.first.as_string_view() == key;
-                                 }));
+  unknown_.erase(std::remove_if(unknown_.begin(), unknown_.end(),
+                                [key](const std::pair<Slice, Slice>& p) {
+                                  return p.first.as_string_view() == key;
+                                }),
+                 unknown_.end());
 }
 
 absl::optional<absl::string_view> UnknownMap::GetStringValue(
@@ -63,7 +125,7 @@ absl::optional<absl::string_view> UnknownMap::GetStringValue(
 }  // namespace metadata_detail
 
 ContentTypeMetadata::MementoType ContentTypeMetadata::ParseMemento(
-    Slice value, MetadataParseErrorFn on_error) {
+    Slice value, bool, MetadataParseErrorFn /*on_error*/) {
   auto out = kInvalid;
   auto value_string = value.as_string_view();
   if (value_string == "application/grpc") {
@@ -75,7 +137,9 @@ ContentTypeMetadata::MementoType ContentTypeMetadata::ParseMemento(
   } else if (value_string.empty()) {
     out = kEmpty;
   } else {
-    on_error("invalid value", value);
+    // We are intentionally not invoking on_error here since the spec is not
+    // clear on what the behavior should be here, so to avoid breaking anyone,
+    // we should continue to accept this.
   }
   return out;
 }
@@ -105,7 +169,7 @@ const char* ContentTypeMetadata::DisplayValue(ValueType content_type) {
 }
 
 GrpcTimeoutMetadata::MementoType GrpcTimeoutMetadata::ParseMemento(
-    Slice value, MetadataParseErrorFn on_error) {
+    Slice value, bool, MetadataParseErrorFn on_error) {
   auto timeout = ParseTimeout(value);
   if (!timeout.has_value()) {
     on_error("invalid value", value);
@@ -127,7 +191,7 @@ Slice GrpcTimeoutMetadata::Encode(ValueType x) {
 }
 
 TeMetadata::MementoType TeMetadata::ParseMemento(
-    Slice value, MetadataParseErrorFn on_error) {
+    Slice value, bool, MetadataParseErrorFn on_error) {
   auto out = kInvalid;
   if (value == "trailers") {
     out = kTrailers;
@@ -179,7 +243,7 @@ size_t EncodedSizeOfKey(HttpSchemeMetadata, HttpSchemeMetadata::ValueType x) {
   }
 }
 
-const char* HttpSchemeMetadata::DisplayValue(MementoType content_type) {
+const char* HttpSchemeMetadata::DisplayValue(ValueType content_type) {
   switch (content_type) {
     case kHttp:
       return "http";
@@ -191,7 +255,7 @@ const char* HttpSchemeMetadata::DisplayValue(MementoType content_type) {
 }
 
 HttpMethodMetadata::MementoType HttpMethodMetadata::ParseMemento(
-    Slice value, MetadataParseErrorFn on_error) {
+    Slice value, bool, MetadataParseErrorFn on_error) {
   auto out = kInvalid;
   auto value_string = value.as_string_view();
   if (value_string == "POST") {
@@ -236,7 +300,7 @@ const char* HttpMethodMetadata::DisplayValue(ValueType content_type) {
 }
 
 CompressionAlgorithmBasedMetadata::MementoType
-CompressionAlgorithmBasedMetadata::ParseMemento(Slice value,
+CompressionAlgorithmBasedMetadata::ParseMemento(Slice value, bool,
                                                 MetadataParseErrorFn on_error) {
   auto algorithm = ParseCompressionAlgorithm(value.as_string_view());
   if (!algorithm.has_value()) {
@@ -247,7 +311,7 @@ CompressionAlgorithmBasedMetadata::ParseMemento(Slice value,
 }
 
 Duration GrpcRetryPushbackMsMetadata::ParseMemento(
-    Slice value, MetadataParseErrorFn on_error) {
+    Slice value, bool, MetadataParseErrorFn on_error) {
   int64_t out;
   if (!absl::SimpleAtoi(value.as_string_view(), &out)) {
     on_error("not an integer", value);
@@ -269,7 +333,7 @@ std::string LbCostBinMetadata::DisplayValue(ValueType x) {
 }
 
 LbCostBinMetadata::MementoType LbCostBinMetadata::ParseMemento(
-    Slice value, MetadataParseErrorFn on_error) {
+    Slice value, bool, MetadataParseErrorFn on_error) {
   if (value.length() < sizeof(double)) {
     on_error("too short", value);
     return {0, ""};
@@ -292,7 +356,14 @@ std::string GrpcStreamNetworkState::DisplayValue(ValueType x) {
   GPR_UNREACHABLE_CODE(return "unknown value");
 }
 
-std::string PeerString::DisplayValue(ValueType x) { return std::string(x); }
+std::string GrpcRegisteredMethod::DisplayValue(void* x) {
+  return absl::StrFormat("%p", x);
+}
+
+std::string PeerString::DisplayValue(const ValueType& x) {
+  return std::string(x.as_string_view());
+}
+
 const std::string& GrpcStatusContext::DisplayValue(const std::string& x) {
   return x;
 }
