@@ -18,6 +18,8 @@
 
 #include "src/core/resolver/xds/xds_dependency_manager.h"
 
+#include <set>
+
 #include "absl/strings/str_join.h"
 
 #include "src/core/ext/xds/xds_routing.h"
@@ -105,7 +107,7 @@ std::string XdsDependencyManager::XdsConfig::ToString() const {
 // XdsDependencyManager::ListenerWatcher
 //
 
-class XdsDependencyManager::ListenerWatcher
+class XdsDependencyManager::ListenerWatcher final
     : public XdsListenerResourceType::WatcherInterface {
  public:
   explicit ListenerWatcher(RefCountedPtr<XdsDependencyManager> dependency_mgr)
@@ -154,7 +156,7 @@ class XdsDependencyManager::ListenerWatcher
 // XdsDependencyManager::RouteConfigWatcher
 //
 
-class XdsDependencyManager::RouteConfigWatcher
+class XdsDependencyManager::RouteConfigWatcher final
     : public XdsRouteConfigResourceType::WatcherInterface {
  public:
   RouteConfigWatcher(RefCountedPtr<XdsDependencyManager> dependency_mgr,
@@ -206,7 +208,7 @@ class XdsDependencyManager::RouteConfigWatcher
 // XdsDependencyManager::ClusterWatcher
 //
 
-class XdsDependencyManager::ClusterWatcher
+class XdsDependencyManager::ClusterWatcher final
     : public XdsClusterResourceType::WatcherInterface {
  public:
   ClusterWatcher(RefCountedPtr<XdsDependencyManager> dependency_mgr,
@@ -255,7 +257,7 @@ class XdsDependencyManager::ClusterWatcher
 // XdsDependencyManager::EndpointWatcher
 //
 
-class XdsDependencyManager::EndpointWatcher
+class XdsDependencyManager::EndpointWatcher final
     : public XdsEndpointResourceType::WatcherInterface {
  public:
   EndpointWatcher(RefCountedPtr<XdsDependencyManager> dependency_mgr,
@@ -306,7 +308,8 @@ class XdsDependencyManager::EndpointWatcher
 // XdsDependencyManager::DnsResultHandler
 //
 
-class XdsDependencyManager::DnsResultHandler : public Resolver::ResultHandler {
+class XdsDependencyManager::DnsResultHandler final
+    : public Resolver::ResultHandler {
  public:
   DnsResultHandler(RefCountedPtr<XdsDependencyManager> dependency_mgr,
                    std::string name)
@@ -330,7 +333,7 @@ class XdsDependencyManager::DnsResultHandler : public Resolver::ResultHandler {
 // XdsDependencyManager::ClusterSubscription
 //
 
-void XdsDependencyManager::ClusterSubscription::Orphan() {
+void XdsDependencyManager::ClusterSubscription::Orphaned() {
   dependency_mgr_->work_serializer_->Run(
       [self = WeakRef()]() {
         self->dependency_mgr_->OnClusterSubscriptionUnref(self->cluster_name_,
@@ -468,7 +471,8 @@ void XdsDependencyManager::OnListenerUpdate(
 
 namespace {
 
-class XdsVirtualHostListIterator : public XdsRouting::VirtualHostListIterator {
+class XdsVirtualHostListIterator final
+    : public XdsRouting::VirtualHostListIterator {
  public:
   explicit XdsVirtualHostListIterator(
       const std::vector<XdsRouteConfigResource::VirtualHost>* virtual_hosts)
@@ -633,11 +637,12 @@ void XdsDependencyManager::OnEndpointUpdate(
     it->second.update.resolution_note =
         absl::StrCat("EDS resource ", name, " contains no localities");
   } else {
-    std::set<std::string> empty_localities;
+    std::set<absl::string_view> empty_localities;
     for (const auto& priority : endpoint->priorities) {
       for (const auto& p : priority.localities) {
         if (p.second.endpoints.empty()) {
-          empty_localities.insert(p.first->AsHumanReadableString());
+          empty_localities.insert(
+              p.first->human_readable_string().as_string_view());
         }
       }
     }
